/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <unistd.h>
#include <stdio.h>

#include <appsvc.h>

#include "unittest.h"
#include "alarm.h"
#include "mock/gio_fake.h"
#include "mock/glib_fake.h"
#include "mock/notification_fake.h"

void __g_dbus_node_info_unref_fake(GDBusNodeInfo *info)
{
  free(info->interfaces);
  free(info);
  info = NULL;
  return;
}

GDBusConnection* __g_bus_get_sync_fake(GBusType type,
    GCancellable *cancellable, GError **error)
{
  GDBusConnection *con = (GDBusConnection*)g_object_new(G_TYPE_OBJECT, NULL);

  return con;
}

GDBusProxy* __g_dbus_proxy_new_sync_fake(GDBusConnection *con,
    GDBusProxyFlags flag, GDBusInterfaceInfo *info, const gchar *name,
    const gchar *path, const gchar *interface, GCancellable *cancellable,
    GError **error)
{
  GDBusProxy *proxy = (GDBusProxy*)g_object_new(G_TYPE_OBJECT, NULL);

  return proxy;
}

GDBusNodeInfo* __g_dbus_node_info_new_for_xml_fake(const gchar *xml, GError **error)
{
  GDBusNodeInfo *info = (GDBusNodeInfo*)malloc(sizeof(GDBusNodeInfo));

  info->ref_count = 10;
  info->path = NULL;
  info->interfaces = (GDBusInterfaceInfo**)malloc(sizeof(GDBusInterfaceInfo));
  info->nodes = NULL;
  info->annotations = NULL;

  return info;
}

guint __g_dbus_connection_register_object_fake(GDBusConnection *con,
    const gchar *path, GDBusInterfaceInfo *info,
    const GDBusInterfaceVTable *table, gpointer user_data,
    GDestroyNotify user_data_free_func, GError **error)
{
  return 1;
}

gboolean __g_dbus_connection_unregister_object_fake(GDBusConnection *con,
    guint id)
{
  return true;
}

guint __g_bus_own_name_on_connection_fake(GDBusConnection *con, const gchar *name,
    GBusNameOwnerFlags flag, GBusNameAcquiredCallback acquired_cb,
    GBusNameLostCallback lost_cb, gpointer user_data,
    GDestroyNotify user_data_free_func)
{
  return 1;
}

GVariant* __g_dbus_proxy_call_sync_fake_negative(GDBusProxy *proxy, const gchar* method,
    GVariant *param, GDBusCallFlags flag, gint timeout, GCancellable *cancellable,
    GError **error)
{
 GQuark quark = g_quark_from_static_string("test");
  *error = g_error_new(quark, G_DBUS_ERROR_ACCESS_DENIED, "Error");

  return NULL;
}

void __g_dbus_proxy_call_fake(GDBusProxy *proxy, const gchar* method,
    GVariant *param, GDBusCallFlags flag, gint timeout, GCancellable *cancellable,
    GAsyncReadyCallback cb, gpointer user_data)
{
  return;
}

gsize __g_variant_get_size_fake(GVariant *var)
{
  int size = 1;
  return (gsize)size;
}

GVariant* __notification_ipc_make_gvariant_from_noti_fake(notification_h,
    bool translate)
{
  GVariant *noti_gv = NULL;

  noti_gv = g_variant_new_string("test");

  return noti_gv;
}

class AlarmLibTest : public ::testing::Test {
  protected:
    void SetUp() override {
      alarm = alarmmgr_create_alarm();

      g_bus_get_sync_fake.custom_fake = __g_bus_get_sync_fake;
      g_dbus_proxy_new_sync_fake.custom_fake = __g_dbus_proxy_new_sync_fake;
      g_dbus_node_info_unref_fake.custom_fake = __g_dbus_node_info_unref_fake;
    }
    void TearDown() override {
      alarmmgr_free_alarm(alarm);
    }

    alarm_entry_t *alarm;
};

TEST_F(AlarmLibTest, alarmmgr_init_p)
{
  int ret;
  char testapp[256] = "org.tizen.alarmmgrtestapp";

  g_dbus_node_info_new_for_xml_fake.custom_fake = __g_dbus_node_info_new_for_xml_fake;
  g_dbus_connection_register_object_fake.custom_fake = __g_dbus_connection_register_object_fake;
  g_dbus_connection_unregister_object_fake.custom_fake = __g_dbus_connection_unregister_object_fake;
  g_bus_own_name_on_connection_fake.custom_fake = __g_bus_own_name_on_connection_fake;

  ret = alarmmgr_init(testapp);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
  alarmmgr_fini();
}

static int __handler(alarm_id_t alarm_id, void *user_param)
{
  return 0;
}

TEST_F(AlarmLibTest, alarmmgr_set_cb_p)
{
  int ret;

  ret = alarmmgr_set_cb(__handler, NULL);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_get_time_p)
{
  int ret;
  alarm_date_t date;
  alarm_date_t get_date;

  date.year = 2020;
  date.month = 1;
  date.day = 1;
  date.hour = 1;
  date.min = 1;
  date.sec = 1;

  ret = alarmmgr_set_time(alarm, date);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_get_time(alarm, &get_date);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  EXPECT_EQ(date.year, get_date.year);
}

TEST_F(AlarmLibTest, alarmmgr_set_get_repeat_mode_p)
{
  int ret;
  alarm_repeat_mode_t get_mode;
  int interval;

  ret = alarmmgr_set_repeat_mode(alarm, ALARM_REPEAT_MODE_REPEAT, 1000);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_get_repeat_mode(alarm, &get_mode, &interval);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  EXPECT_EQ(ALARM_REPEAT_MODE_REPEAT, get_mode);
}

TEST_F(AlarmLibTest, alarmmgr_set_get_type_p)
{
  int ret;
  int get_type;

  ret = alarmmgr_set_type(alarm, ALARM_TYPE_VOLATILE);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_get_type(alarm, &get_type);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  EXPECT_EQ(ALARM_TYPE_VOLATILE, get_type);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_appsvc_n)
{
  int ret;
  alarm_id_t alarm_id;

  bundle *b;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  b = bundle_create();
  ASSERT_NE(b, nullptr);

  ret = appsvc_set_pkgname(b,"org.tizen.calendar");
  EXPECT_EQ(0, ret);

  ret = appsvc_set_operation(b, APPSVC_OPERATION_DEFAULT);
  EXPECT_EQ(0, ret);

  ret = alarmmgr_add_alarm_appsvc(ALARM_TYPE_DEFAULT, 100, 0, (void *)b,
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
  bundle_free(b);
}

TEST_F(AlarmLibTest, alarmmgr_get_alarm_appsvc_n)
{
  int ret;
  alarm_id_t alarm_id = 10;

  bundle *get_b;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  get_b = (bundle *)alarmmgr_get_alarm_appsvc_info(alarm_id, &ret);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);

  bundle_free(get_b);
}


TEST_F(AlarmLibTest, alarmmgr_add_alarm_noti_n)
{
  int ret;
  alarm_id_t alarm_id;
  notification_h noti = nullptr;

  g_variant_get_size_fake.custom_fake = __g_variant_get_size_fake;
  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;
  notification_ipc_make_gvariant_from_noti_fake.custom_fake = __notification_ipc_make_gvariant_from_noti_fake;

  ret = alarmmgr_add_alarm_noti(ALARM_TYPE_DEFAULT, 100, 0, noti,
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}


TEST_F(AlarmLibTest, alarmmgr_get_alarm_noti_n)
{
  int ret;
  notification_h get_noti;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_get_alarm_noti_info(1, &get_noti);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_rtc_time_n)
{
  int ret;
  alarm_date_t date = {2020, 4, 5, 10, 10, 0};

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_set_rtc_time(&date);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_with_localtime_n)
{
  int ret;
  alarm_id_t alarm_id;
  alarm_date_t test_time;
  char testapp[256] = "org.tizen.alarmmgrtestapp";

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  test_time.year = 2022;
  test_time.month = 1;
  test_time.day = 1;

  test_time.hour = 1;
  test_time.min = 1;
  test_time.sec = 5;

  ret = alarmmgr_init(testapp);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_set_time(alarm,test_time);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_set_type(alarm, ALARM_TYPE_DEFAULT);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_set_repeat_mode(alarm, ALARM_REPEAT_MODE_REPEAT, 1000);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_add_alarm_with_localtime(alarm, testapp, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
  alarmmgr_fini();
}

TEST_F(AlarmLibTest, alarmmgr_get_info_n)
{
  int ret;
  alarm_entry_t get_alarm;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_get_info(1, &get_alarm);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_appsvc_with_localtime_n)
{
  int ret;
  alarm_id_t alarm_id;

  bundle *b;

  alarm_date_t test_time;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  b = bundle_create();
  ASSERT_NE(b, nullptr);

  ret = appsvc_set_pkgname(b,"org.tizen.calendar");
  EXPECT_EQ(0, ret);

  ret = appsvc_set_operation(b, APPSVC_OPERATION_DEFAULT);
  EXPECT_EQ(0, ret);

  test_time.year = 2022;
  test_time.month = 1;
  test_time.day = 1;

  test_time.hour = 1;
  test_time.min = 1;
  test_time.sec = 5;

  ret = alarmmgr_set_time(alarm,test_time);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_set_type(alarm, ALARM_TYPE_DEFAULT);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_add_alarm_appsvc_with_localtime(alarm, (void *)b,
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);

  bundle_free(b);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_noti_with_localtime_n)
{
  int ret;
  alarm_id_t alarm_id;
  notification_h noti = nullptr;

  notification_ipc_make_gvariant_from_noti_fake.custom_fake = __notification_ipc_make_gvariant_from_noti_fake;
  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_set_repeat_mode(alarm, ALARM_REPEAT_MODE_REPEAT, 1000);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_add_alarm_noti_with_localtime(alarm, noti, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_n)
{
  int ret;
  alarm_id_t alarm_id;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_set_repeat_mode(alarm, ALARM_REPEAT_MODE_REPEAT, 1000);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_add_alarm(ALARM_TYPE_VOLATILE, 100, 1000, "gtest-alarmmgr",
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_precision_n)
{
  int ret;
  alarm_id_t alarm_id;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_add_alarm_precision(ALARM_TYPE_VOLATILE, 100, 1000,
      "gtest-alarmmgr", &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_get_alarm_withcb_n)
{
  int ret;
  alarm_id_t alarm_id;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_add_alarm_withcb(ALARM_TYPE_VOLATILE, 100, 1000,
      __handler, NULL, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_get_alarm_withcb_precision_n)
{
  int ret;
  alarm_id_t alarm_id;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_add_alarm_withcb_precision(ALARM_TYPE_VOLATILE, 100, 1000,
      __handler, NULL, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_remove_alarm_n)
{
  int ret;
  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_remove_alarm(1);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_remove_all_n)
{
  int ret;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_remove_all();
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

int enum_fn(alarm_id_t alarm_id, void *user_param)
{
  return 0;
}

TEST_F(AlarmLibTest, alarmmgr_enum_alarm_ids_n)
{
  int ret;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_enum_alarm_ids(enum_fn, nullptr);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_get_next_duetime_n)
{
  int ret;
  time_t next_time;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_get_next_duetime(1, &next_time);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

int handler(alarm_id_t alarm_id, void *user_param)
{
  return 0;
}

TEST_F(AlarmLibTest, alarmmgr_add_periodic_alarm_withcb_n)
{
  int ret;
  alarm_id_t alarm_id;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_add_periodic_alarm_withcb(1000, CUT_OFF, handler, NULL,
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_reference_periodic_alarm_withcb_n)
{
  int ret;
  alarm_id_t alarm_id;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_add_reference_periodic_alarm_withcb(1000, handler, NULL,
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_systime_n)
{
  int ret;
  time_t current_time;

  time(&current_time);

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_set_systime(current_time);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_systime_with_propagation_delay_n)
{
  int ret;
  struct timespec newtime;
  struct timespec reqtime;

  clock_gettime(CLOCK_REALTIME, &newtime);
  clock_gettime(CLOCK_REALTIME, &reqtime);

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_set_systime_with_propagation_delay(newtime, reqtime);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

int set_time_cb(int result, void *user_param)
{
  return 0;
}

TEST_F(AlarmLibTest, alarmmgr_set_systime_async_p)
{
  int ret;
  time_t current_time;

  g_dbus_proxy_call_fake.custom_fake = __g_dbus_proxy_call_fake;

  time(&current_time);

  ret = alarmmgr_set_systime_async(current_time, set_time_cb, NULL);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_systime_with_propagation_delay_async_p)
{
  int ret;
  struct timespec newtime;
  struct timespec reqtime;

  g_dbus_proxy_call_fake.custom_fake = __g_dbus_proxy_call_fake;

  clock_gettime(CLOCK_REALTIME, &newtime);
  clock_gettime(CLOCK_REALTIME, &reqtime);

  newtime.tv_sec += 100;
  reqtime.tv_sec += 100;

  ret = alarmmgr_set_systime_with_propagation_delay_async(newtime,
      reqtime, set_time_cb, NULL);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_timezone_n)
{
  int ret;
  char zone[] = "/usr/share/zoneinfo/Asia/Seoul";

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_set_timezone(zone);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_global_n)
{
  int ret;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_set_global(1, true);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_get_global_n)
{
  int ret;
  bool global;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  ret = alarmmgr_get_global(1, &global);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_update_alarm_n)
{
  int ret;
  alarm_date_t test_time;

  g_dbus_proxy_call_sync_fake.custom_fake = __g_dbus_proxy_call_sync_fake_negative;

  test_time.year = 2022;
  test_time.month = 1;
  test_time.day = 1;

  test_time.hour = 1;
  test_time.min = 1;
  test_time.sec = 5;

  ret = alarmmgr_set_time(alarm, test_time);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_update_alarm(1, alarm, ALARM_UPDATE_FLAG_TIME);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}
